@@ -1,9 +1,9 @@
 /*
-Copyright (c) 2013, Michael Ossmann <mike@ossmann.com>
-Copyright (c) 2012, Jared Boone <jared@sharebrained.com>
-Copyright (c) 2013-2017, Youssef Touil <youssef@airspy.com>
+Copyright (c) 2013-2018, Youssef Touil <youssef@airspy.com>
 Copyright (c) 2013-2017, Ian Gilmour <ian@sdrsharp.com>
 Copyright (c) 2013-2017, Benjamin Vernoux <bvernoux@airspy.com>
+Copyright (c) 2013, Michael Ossmann <mike@ossmann.com>
+Copyright (c) 2012, Jared Boone <jared@sharebrained.com>
 
 All rights reserved.
 
@@ -45,12 +45,14 @@ typedef int bool;
 #define false 0
 #endif
 
+#ifndef M_PI
 #define M_PI (3.14159265359)
+#endif
 
 #define MAX(a,b) ((a) > (b) ? a : b)
 #define MIN(a,b) ((a) < (b) ? a : b)
 
-#define SAMPLES_TO_TRANSFER (1024 * 16)
+#define SAMPLES_TO_TRANSFER (1024 * 2)
 #define SERIAL_NUMBER_UNUSED (0)
 #define RAW_BUFFER_COUNT (8)
 #define AIRSPYHF_SERIAL_SIZE (28)
@@ -60,20 +62,16 @@ typedef int bool;
 
 #define CALIBRATION_MAGIC (0xA5CA71B0)
 
-#define ERASE_FLASH_MAGIC0 (0xC0DE)
-#define ERASE_FLASH_MAGIC1 (0xBABE)
+#define DEFAULT_IF_SHIFT (5000)
+#define MIN_ZERO_IF_LO (180)
+#define MIN_LOW_IF_LO (84)
 
 #define STR_PREFIX_SERIAL_AIRSPYHF_SIZE (12)
 static const char str_prefix_serial_airspyhf[STR_PREFIX_SERIAL_AIRSPYHF_SIZE] =
 { 'A', 'I', 'R', 'S', 'P', 'Y', 'H', 'F', ' ', 'S', 'N', ':' };
 
-#define VERSION_STRING_SIZE (64)
-
-#ifdef AIRSPYHF_BIG_ENDIAN
-#define TO_LE(x) __builtin_bswap32(x)
-#else
-#define TO_LE(x) x
-#endif
+#define INITIAL_PHASE (0.00006f)
+#define INITIAL_AMPLITUDE (-0.0045f)
 
 #pragma pack(push,1)
 
@@ -84,7 +82,7 @@ typedef struct {
 
 #pragma pack(pop)
 
-struct airspyhf_device
+typedef struct airspyhf_device
 {
 	libusb_context* usb_context;
 	libusb_device_handle* usb_device;
@@ -96,13 +94,18 @@ struct airspyhf_device
 	pthread_mutex_t consumer_mp;
 	uint32_t supported_samplerate_count;
 	uint32_t *supported_samplerates;
+	uint8_t *samplerate_architectures;
 	volatile uint32_t current_samplerate;
 	volatile uint32_t freq_hz;
 	volatile uint32_t freq_khz;
 	volatile int32_t freq_shift;
 	volatile int32_t calibration_ppb;
+	volatile float optimal_point;
+	uint8_t enable_dsp;
+	uint8_t is_low_if;
+	float filter_gain;
 	airspyhf_complex_float_t vec;
-	iq_balancer_t iq_balancer;
+	struct iq_balancer_t *iq_balancer;
 	uint32_t transfer_count;
 	uint32_t buffer_size;
 	uint32_t dropped_buffers;
@@ -115,7 +118,7 @@ struct airspyhf_device
 	volatile int received_buffer_count;
 	airspyhf_complex_float_t *output_buffer;
 	void* ctx;
-};
+} airspyhf_device_t;
 
 typedef struct calibration_record
 {
@@ -292,28 +295,46 @@ static void convert_samples(airspyhf_device_t* device, airspyhf_complex_int16_t
 	const float scale = 1.0f / 32768;
 
 	int i;
-	airspyhf_complex_float_t vec = device->vec;
+	airspyhf_complex_float_t vec;
 	airspyhf_complex_float_t rot;
-	double angle = 2.0 * M_PI * device->freq_shift / (double) device->current_samplerate;
+	double angle;
+	float conversion_gain;
 
-	rot.re = (float) cos(angle);
-	rot.im = (float) -sin(angle);
+	conversion_gain = scale * device->filter_gain;
 
 	for (i = 0; i < count; i++)
 	{
-		dest[i].re = src[i].re * scale;
-		dest[i].im = src[i].im * scale;
+		dest[i].re = src[i].re * conversion_gain;
+		dest[i].im = src[i].im * conversion_gain;
 	}
 
-	iq_balancer_process(&device->iq_balancer, dest, count);
-
-	for (i = 0; i < count; i++)
+	if (device->enable_dsp)
 	{
-		rotate_complex(&vec, &rot);
-		multiply_complex_complex(&dest[i], &vec);
-	}
+		if (!device->is_low_if)
+		{
+			// Zero IF requires external IQ correction
+			iq_balancer_process(device->iq_balancer, dest, count);
+		}
 
-	device->vec = vec;
+		// Fine tuning
+		if (device->freq_shift != 0)
+		{
+			angle = 2.0 * M_PI * device->freq_shift / (double) device->current_samplerate;
+
+			vec = device->vec;
+
+			rot.re = (float) cos(angle);
+			rot.im = (float) -sin(angle);
+
+			for (i = 0; i < count; i++)
+			{
+				rotate_complex(&vec, &rot);
+				multiply_complex_complex(&dest[i], &vec);
+			}
+
+			device->vec = vec;
+		}
+	}
 }
 
 static void* consumer_threadproc(void *arg)
@@ -525,19 +546,6 @@ static void airspyhf_open_exit(airspyhf_device_t* device)
 	device->usb_context = NULL;
 }
 
-static void upper_string(unsigned char *string, size_t len)
-{
-	while (len > 0)
-	{
-		if (*string >= 'a' && *string <= 'z')
-		{
-			*string = *string - 32;
-		}
-		string++;
-		len--;
-	}
-}
-
 static int airspyhf_read_samplerates_from_fw(airspyhf_device_t* device, uint32_t* buffer, const uint32_t len)
 {
 	int result;
@@ -559,6 +567,28 @@ static int airspyhf_read_samplerates_from_fw(airspyhf_device_t* device, uint32_t
 	return AIRSPYHF_SUCCESS;
 }
 
+static int airspyhf_read_samplerate_architectures_from_fw(airspyhf_device_t* device, uint8_t* buffer, const uint32_t len)
+{
+	int result;
+
+	result = libusb_control_transfer(
+		device->usb_device,
+		LIBUSB_ENDPOINT_IN | LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_RECIPIENT_DEVICE,
+		AIRSPYHF_GET_SAMPLERATE_ARCHITECTURES,
+		0,
+		len,
+		(unsigned char*) buffer,
+		(len > 0 ? len : 1) * (int16_t) sizeof(uint32_t),
+		0);
+
+	if (result < 1)
+	{
+		return AIRSPYHF_ERROR;
+	}
+
+	return AIRSPYHF_SUCCESS;
+}
+
 
 static void airspyhf_open_device(airspyhf_device_t* device,
 	int* ret,
@@ -577,10 +607,7 @@ static void airspyhf_open_device(airspyhf_device_t* device,
 	ssize_t cnt;
 	int serial_descriptor_index;
 	struct libusb_device_descriptor device_descriptor;
-	char serial_number_expected[AIRSPYHF_SERIAL_SIZE + 1];
 	unsigned char serial_number[AIRSPYHF_SERIAL_SIZE + 1];
-	uint32_t serial_number_msb_val;
-	uint32_t serial_number_lsb_val;
 
 	libusb_dev_handle = &device->usb_device;
 	*libusb_dev_handle = NULL;
@@ -615,19 +642,19 @@ static void airspyhf_open_device(airspyhf_device_t* device,
 						serial_descriptor_index,
 						serial_number,
 						sizeof(serial_number));
-					if (serial_number_len == AIRSPYHF_SERIAL_SIZE)
+
+					if (serial_number_len == AIRSPYHF_SERIAL_SIZE && !memcmp(str_prefix_serial_airspyhf, serial_number, STR_PREFIX_SERIAL_AIRSPYHF_SIZE))
 					{
-						serial_number[AIRSPYHF_SERIAL_SIZE] = 0;
-						upper_string(serial_number, AIRSPYHF_SERIAL_SIZE);
-						serial_number_msb_val = (uint32_t)(serial_number_val >> 32);
-						serial_number_lsb_val = (uint32_t)(serial_number_val & 0xFFFFFFFF);
-
-						sprintf(serial_number_expected, "%s%08X%08X",
-							str_prefix_serial_airspyhf,
-							serial_number_msb_val,
-							serial_number_lsb_val);
-
-						if (strncmp((const char*)serial_number, serial_number_expected, AIRSPYHF_SERIAL_SIZE) == 0)
+						uint64_t serial = SERIAL_NUMBER_UNUSED;
+						// use same code to detemine device's serial number as in airspyhf_list_devices()
+						{
+							char *start, *end;
+							serial_number[AIRSPYHF_SERIAL_SIZE] = 0;
+							start = (char*)(serial_number + STR_PREFIX_SERIAL_AIRSPYHF_SIZE);
+							end = NULL;
+							serial = strtoull(start, &end, 16);
+						}
+						if (serial == serial_number_val)
 						{
 							result = libusb_set_configuration(dev_handle, 1);
 							if (result != 0)
@@ -728,7 +755,10 @@ int airspyhf_list_devices(uint64_t *serials, int count)
 	int i;
 	unsigned char serial_number[AIRSPYHF_SERIAL_SIZE + 1];
 
-	memset(serials, 0, sizeof(uint64_t) * count);
+	if (serials)
+	{
+		memset(serials, 0, sizeof(uint64_t) * count);
+	}
 
 	if (libusb_init(&context) != 0)
 	{
@@ -742,7 +772,7 @@ int airspyhf_list_devices(uint64_t *serials, int count)
 
 	i = 0;
 	output_count = 0;
-	while ((dev = devices[i++]) != NULL && output_count < count)
+	while ((dev = devices[i++]) != NULL && (!serials || output_count < count))
 	{
 		libusb_get_device_descriptor(dev, &device_descriptor);
 
@@ -762,7 +792,7 @@ int airspyhf_list_devices(uint64_t *serials, int count)
 					serial_number,
 					sizeof(serial_number));
 
-				if (serial_number_len == AIRSPYHF_SERIAL_SIZE)
+				if (serial_number_len == AIRSPYHF_SERIAL_SIZE && !memcmp(str_prefix_serial_airspyhf, serial_number, STR_PREFIX_SERIAL_AIRSPYHF_SIZE))
 				{
 					char *start, *end;
 					uint64_t serial;
@@ -777,7 +807,10 @@ int airspyhf_list_devices(uint64_t *serials, int count)
 						continue;
 					}
 
-					serials[output_count] = serial;
+					if (serials)
+					{
+						serials[output_count] = serial;
+					}
 					output_count++;
 				}
 
@@ -833,12 +866,25 @@ static int airspyhf_open_init(airspyhf_device_t** device, uint64_t serial_number
 	lib_device->streaming = false;
 	lib_device->stop_requested = false;
 
+	lib_device->supported_samplerates = NULL;
+	lib_device->samplerate_architectures = NULL;
+
 	result = airspyhf_read_samplerates_from_fw(lib_device, &lib_device->supported_samplerate_count, 0);
 	if (result == AIRSPYHF_SUCCESS)
 	{
 		lib_device->supported_samplerates = (uint32_t *)malloc(lib_device->supported_samplerate_count * sizeof(uint32_t));
 		result = airspyhf_read_samplerates_from_fw(lib_device, lib_device->supported_samplerates, lib_device->supported_samplerate_count);
-		if (result != AIRSPYHF_SUCCESS)
+		if (result == AIRSPYHF_SUCCESS)
+		{
+			lib_device->samplerate_architectures = (uint8_t *)malloc(lib_device->supported_samplerate_count * sizeof(uint8_t));
+			result = airspyhf_read_samplerate_architectures_from_fw(lib_device, lib_device->samplerate_architectures, lib_device->supported_samplerate_count);
+			if (result != AIRSPYHF_SUCCESS)
+			{
+				memset(lib_device->samplerate_architectures, 0, lib_device->supported_samplerate_count * sizeof(uint8_t)); // Asume Zero IF for all
+				result = AIRSPYHF_SUCCESS; // Clear this error for backward compatibility.
+			}
+		}
+		else
 		{
 			free(lib_device->supported_samplerates);
 			lib_device->supported_samplerates = NULL;
@@ -847,12 +893,18 @@ static int airspyhf_open_init(airspyhf_device_t** device, uint64_t serial_number
 
 	if (result != AIRSPYHF_SUCCESS)
 	{
+		// Assume one default sample rate with Zero IF
+
 		lib_device->supported_samplerate_count = 1;
 		lib_device->supported_samplerates = (uint32_t *) malloc(lib_device->supported_samplerate_count * sizeof(uint32_t));
 		lib_device->supported_samplerates[0] = DEFAULT_SAMPLERATE;
+
+		lib_device->samplerate_architectures = (uint8_t *) malloc(lib_device->supported_samplerate_count * sizeof(uint8_t));
+		lib_device->samplerate_architectures[0] = 0;
 	}
 
 	lib_device->current_samplerate = lib_device->supported_samplerates[0];
+	lib_device->is_low_if = lib_device->samplerate_architectures[0];
 	
 	result = allocate_transfers(lib_device);
 	if (result != 0)
@@ -870,6 +922,9 @@ static int airspyhf_open_init(airspyhf_device_t** device, uint64_t serial_number
 	lib_device->freq_shift = 0;
 	lib_device->vec.re = 1.0f;
 	lib_device->vec.im = 0.0f;
+	lib_device->optimal_point = 0.0f;
+	lib_device->filter_gain = 1.0f;
+	lib_device->enable_dsp = 1;
 
 	if (airspyhf_config_read(lib_device, (uint8_t *) &record, sizeof(record)) == AIRSPYHF_SUCCESS)
 	{
@@ -887,7 +942,7 @@ static int airspyhf_open_init(airspyhf_device_t** device, uint64_t serial_number
 		lib_device->calibration_ppb = 0;
 	}
 
-	iq_balancer_init(&lib_device->iq_balancer);
+	lib_device->iq_balancer = iq_balancer_create(INITIAL_PHASE, INITIAL_AMPLITUDE);
 
 	*device = lib_device;
 
@@ -933,12 +988,25 @@ int ADDCALL airspyhf_close(airspyhf_device_t* device)
 		airspyhf_open_exit(device);
 		free_transfers(device);
 		free(device->supported_samplerates);
+		free(device->samplerate_architectures);
+		iq_balancer_destroy(device->iq_balancer);
 		free(device);
 	}
 
 	return result;
 }
 
+int ADDCALL airspyhf_get_output_size(airspyhf_device_t * device)
+{
+	// Todo: Make this configurable
+	return SAMPLES_TO_TRANSFER;
+}
+
+int ADDCALL airspyhf_is_low_if(airspyhf_device_t* device)
+{
+	return device->is_low_if;
+}
+
 int ADDCALL airspyhf_get_samplerates(airspyhf_device_t* device, uint32_t* buffer, const uint32_t len)
 {
 	if (len == 0)
@@ -961,6 +1029,7 @@ int ADDCALL airspyhf_set_samplerate(airspyhf_device_t* device, uint32_t samplera
 {
 	int result;
 	uint32_t i;
+	uint8_t gain;
 
 	if (samplerate > MAX_SAMPLERATE_INDEX)
 	{
@@ -979,8 +1048,37 @@ int ADDCALL airspyhf_set_samplerate(airspyhf_device_t* device, uint32_t samplera
 		return AIRSPYHF_ERROR;
 	}
 
+	device->current_samplerate = device->supported_samplerates[samplerate];
+	device->is_low_if = device->samplerate_architectures[samplerate];
+
 	libusb_clear_halt(device->usb_device, LIBUSB_ENDPOINT_IN | 1);
 
+	if (!device->is_low_if && device->freq_khz < MIN_ZERO_IF_LO)
+	{
+		uint8_t buf[4];
+		buf[0] = (uint8_t)((MIN_ZERO_IF_LO >> 24) & 0xff);
+		buf[1] = (uint8_t)((MIN_ZERO_IF_LO >> 16) & 0xff);
+		buf[2] = (uint8_t)((MIN_ZERO_IF_LO >> 8) & 0xff);
+		buf[3] = (uint8_t)((MIN_ZERO_IF_LO) & 0xff);
+
+		result = libusb_control_transfer(
+			device->usb_device,
+			LIBUSB_ENDPOINT_OUT | LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_RECIPIENT_DEVICE,
+			AIRSPYHF_SET_FREQ,
+			0,
+			0,
+			(unsigned char*) &buf,
+			sizeof(buf),
+			0);
+
+		if (result < sizeof(buf))
+		{
+			return AIRSPYHF_ERROR;
+		}
+
+		device->freq_khz = MIN_ZERO_IF_LO;
+	}
+
 	result = libusb_control_transfer(
 		device->usb_device,
 		LIBUSB_ENDPOINT_OUT | LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_RECIPIENT_DEVICE,
@@ -997,7 +1095,26 @@ int ADDCALL airspyhf_set_samplerate(airspyhf_device_t* device, uint32_t samplera
 		return AIRSPYHF_ERROR;
 	}
 
-	device->current_samplerate = device->supported_samplerates[samplerate];
+	result = libusb_control_transfer(
+		device->usb_device,
+		LIBUSB_ENDPOINT_IN | LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_RECIPIENT_DEVICE,
+		AIRSPYHF_GET_FILTER_GAIN,
+		0,
+		0,
+		&gain,
+		sizeof(uint8_t),
+		0);
+
+	if (result == sizeof(uint8_t))
+	{
+		device->filter_gain = powf(10.0f, (float) gain * -0.05f);
+	}
+	else
+	{
+		device->filter_gain = 1.0;
+	}
+
+	airspyhf_set_freq(device, device->freq_hz);
 	
 	return AIRSPYHF_SUCCESS;
 }
@@ -1032,7 +1149,6 @@ int ADDCALL airspyhf_start(airspyhf_device_t* device, airspyhf_sample_block_cb_f
 
 	device->vec.re = 1.0f;
 	device->vec.im = 0.0f;
-	iq_balancer_init(&device->iq_balancer);
 
 	result = airspyhf_set_receiver_mode(device, RECEIVER_MODE_OFF);
 	if (result != AIRSPYHF_SUCCESS)
@@ -1062,6 +1178,7 @@ int ADDCALL airspyhf_stop(airspyhf_device_t* device)
 	int result1, result2;
 	result1 = kill_io_threads(device);
 	result2 = airspyhf_set_receiver_mode(device, RECEIVER_MODE_OFF);
+
 	if (result2 != AIRSPYHF_SUCCESS)
 	{
 		return result2;
@@ -1072,15 +1189,14 @@ int ADDCALL airspyhf_stop(airspyhf_device_t* device)
 int ADDCALL airspyhf_set_freq(airspyhf_device_t* device, const uint32_t freq_hz)
 {
 	const int tuning_alignment = 1000;
-	const int if_shift = 5000;
-	const uint32_t lo_low_khz = 150;
 
 	int result;
 	uint8_t buf[4];
-	
-	uint32_t adjusted_freq_hz = (uint32_t) ((int64_t) freq_hz * (int64_t) (1000000000LL + device->calibration_ppb) / 1000000000LL);
+	uint32_t lo_low_khz = device->is_low_if ? MIN_LOW_IF_LO : MIN_ZERO_IF_LO;
+	uint32_t if_shift = (device->enable_dsp && !device->is_low_if) ? DEFAULT_IF_SHIFT : 0;
+	uint32_t adjusted_freq_hz = (uint32_t) ((int64_t) freq_hz * (int64_t)(1000000000LL + device->calibration_ppb) / 1000000000LL);
 	uint32_t freq_khz = MAX(lo_low_khz, (adjusted_freq_hz + if_shift + tuning_alignment / 2) / tuning_alignment);
-	
+
 	if (device->freq_khz != freq_khz)
 	{
 		buf[0] = (uint8_t) ((freq_khz >> 24) & 0xff);
@@ -1094,10 +1210,12 @@ int ADDCALL airspyhf_set_freq(airspyhf_device_t* device, const uint32_t freq_hz)
 			AIRSPYHF_SET_FREQ,
 			0,
 			0,
-			(unsigned char*)&buf,
+			(unsigned char*) &buf,
 			sizeof(buf),
 			0);
 
+		iq_balancer_set_optimal_point(device->iq_balancer, device->optimal_point);
+
 		if (result < sizeof(buf))
 		{
 			return AIRSPYHF_ERROR;
@@ -1181,6 +1299,11 @@ int ADDCALL airspyhf_flash_calibration(airspyhf_device_t* device)
 	record.magic_number = CALIBRATION_MAGIC;
 	record.calibration_ppb = device->calibration_ppb;
 
+	if (airspyhf_is_streaming(device))
+	{
+		return AIRSPYHF_ERROR;
+	}
+
 	if (airspyhf_config_write(device, (uint8_t *)&record, sizeof(record)) != AIRSPYHF_SUCCESS)
 	{
 		return AIRSPYHF_ERROR;
@@ -1195,6 +1318,19 @@ int ADDCALL airspyhf_set_calibration(airspyhf_device_t* device, int32_t ppb)
 	return airspyhf_set_freq(device, device->freq_hz);
 }
 
+int ADDCALL airspyhf_set_optimal_iq_correction_point(airspyhf_device_t* device, float w)
+{
+	device->optimal_point = w;
+	iq_balancer_set_optimal_point(device->iq_balancer, device->optimal_point);
+	return AIRSPYHF_SUCCESS;
+}
+
+int ADDCALL airspyhf_iq_balancer_configure(airspyhf_device_t* device, int buffers_to_skip, int fft_integration, int fft_overlap, int correlation_integration)
+{
+	iq_balancer_configure(device->iq_balancer, buffers_to_skip, fft_integration, fft_overlap, correlation_integration);
+	return AIRSPYHF_SUCCESS;
+}
+
 int ADDCALL airspyhf_board_partid_serialno_read(airspyhf_device_t* device, airspyhf_read_partid_serialno_t* read_partid_serialno)
 {
 	uint8_t length;
@@ -1222,7 +1358,7 @@ int ADDCALL airspyhf_board_partid_serialno_read(airspyhf_device_t* device, airsp
 int ADDCALL airspyhf_version_string_read(airspyhf_device_t* device, char* version, uint8_t length)
 {
 	int result;
-	char version_local[VERSION_STRING_SIZE];
+	char version_local[MAX_VERSION_STRING_SIZE];
 
 	result = libusb_control_transfer(
 		device->usb_device,
@@ -1230,8 +1366,8 @@ int ADDCALL airspyhf_version_string_read(airspyhf_device_t* device, char* versio
 		AIRSPYHF_GET_VERSION_STRING,
 		0,
 		0,
-		(unsigned char*)version_local,
-		(VERSION_STRING_SIZE - 1),
+		(unsigned char*) version_local,
+		(MAX_VERSION_STRING_SIZE - 1),
 		0);
 
 	if (result < 0)
@@ -1273,4 +1409,100 @@ int ADDCALL airspyhf_set_user_output(airspyhf_device_t* device, airspyhf_user_ou
 	}
 
 	return AIRSPYHF_SUCCESS;
-}
\ No newline at end of file
+}
+
+int ADDCALL airspyhf_set_hf_agc(airspyhf_device_t* device, uint8_t flag)
+{
+	int result;
+
+	result = libusb_control_transfer(
+		device->usb_device,
+		LIBUSB_ENDPOINT_OUT | LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_RECIPIENT_DEVICE,
+		AIRSPYHF_SET_HF_AGC,
+		(uint16_t) flag,
+		0,
+		NULL,
+		0,
+		0);
+
+	if (result < 0)
+	{
+		return AIRSPYHF_ERROR;
+	}
+
+	return AIRSPYHF_SUCCESS;
+}
+
+int ADDCALL airspyhf_set_hf_agc_threshold(airspyhf_device_t* device, uint8_t flag)
+{
+	int result;
+
+	result = libusb_control_transfer(
+		device->usb_device,
+		LIBUSB_ENDPOINT_OUT | LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_RECIPIENT_DEVICE,
+		AIRSPYHF_SET_HF_AGC_THRESHOLD,
+		(uint16_t) flag,
+		0,
+		NULL,
+		0,
+		0);
+
+	if (result < 0)
+	{
+		return AIRSPYHF_ERROR;
+	}
+
+	return AIRSPYHF_SUCCESS;
+}
+
+int ADDCALL airspyhf_set_hf_att(airspyhf_device_t* device, uint8_t value)
+{
+	int result;
+
+	result = libusb_control_transfer(
+		device->usb_device,
+		LIBUSB_ENDPOINT_OUT | LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_RECIPIENT_DEVICE,
+		AIRSPYHF_SET_HF_ATT,
+		(uint16_t) value,
+		0,
+		NULL,
+		0,
+		0);
+
+	if (result < 0)
+	{
+		return AIRSPYHF_ERROR;
+	}
+
+	return AIRSPYHF_SUCCESS;
+}
+
+int ADDCALL airspyhf_set_hf_lna(airspyhf_device_t* device, uint8_t flag)
+{
+	int result;
+
+	result = libusb_control_transfer(
+		device->usb_device,
+		LIBUSB_ENDPOINT_OUT | LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_RECIPIENT_DEVICE,
+		AIRSPYHF_SET_HF_LNA,
+		(uint16_t) flag,
+		0,
+		NULL,
+		0,
+		0);
+
+	if (result < 0)
+	{
+		return AIRSPYHF_ERROR;
+	}
+
+	return AIRSPYHF_SUCCESS;
+}
+
+int ADDCALL airspyhf_set_lib_dsp(airspyhf_device_t* device, const uint8_t flag)
+{
+	device->enable_dsp = flag;
+	return AIRSPYHF_SUCCESS;
+}
+
+//#include "special.c"
